package main;

use strict;
use utf8;
use File::Basename;
use Getopt::Long qw(:config no_ignore_case);

use constant OK => 0;
use constant WARNING => 1;
use constant CRITICAL => 2;
use constant UNKNOWN => 3;

my($opt_f, $opt_V, $opt_h, $opt_t, $opt_v, $opt_d);
my($opt_tag, $opt_logfile, $opt_rotation, $opt_okpattern,
    $opt_criticalpattern,$opt_criticalexception,
    $opt_warningpattern, $opt_warningexception, 
    $opt_noprotocol, $opt_syslogserver, $opt_syslogclient,
    $opt_noperfdata, $opt_sticky,
    $opt_selectedsearches, $opt_template);

my $plugin_revision = '$Revision: 1.0 $ ';
my $progname = basename($0);

sub print_version {
  printf "%s v#PACKAGE_VERSION#\n", basename($0);
}

sub print_help {
  print <<EOTXT;
This Nagios Plugin comes with absolutely NO WARRANTY. You may use
it on your own risk!
Copyright by ConSol Software GmbH, Gerhard Lausser.

This plugin looks for patterns in logfiles, even in those who were rotated
since the last run of this plugin.

Usage: check_logfiles [-t timeout] -f <configfile>

The configfile looks like this:

\$seekfilesdir = '/opt/nagios/var/tmp';
# where the state information will be saved.

\$protocolsdir = '/opt/nagios/var/tmp';
# where protocols with found patterns will be stored.

\$scriptpath = '/opt/nagios/var/tmp';
# where scripts will be searched for.

\$MACROS = \{ CL_DISK01 => "/dev/dsk/c0d1", CL_DISK02 => "/dev/dsk/c0d2" \};

\@searches = (
  {
    tag => 'temperature',
    logfile => '/var/adm/syslog/syslog.log',
    rotation => 'bmwhpux',
    criticalpatterns => ['OVERTEMP_EMERG', 'Power supply failed'],
    warningpatterns => ['OVERTEMP_CRIT', 'Corrected ECC Error'],
    options => 'script,protocol,nocount',
    script => 'sendnsca_cmd'
  },
  {
    tag => 'scsi',
    logfile => '/var/adm/messages',
    rotation => 'solaris',
    criticalpatterns => 'Sense Key: Not Ready',
    criticalexceptions => 'Sense Key: Not Ready /dev/testdisk',
    options => 'noprotocol'
  },
  {
    tag => 'logins',
    logfile => '/var/adm/messages',
    rotation => 'solaris',
    criticalpatterns => ['illegal key', 'read error.*\$CL_DISK01\$'],
    criticalthreshold => 4
    warningpatterns => ['read error.*\$CL_DISK02\$'],
  }
);

EOTXT
}

sub print_usage {
  print <<EOTXT;
Usage: check_logfiles [-t timeout] -f <configfile> [--searches=tag1,tag2,...]
       check_logfiles [-t timeout] --logfile=<logfile> --tag=<tag> --rotation=<rotation>
                      --criticalpattern=<regexp> --warningpattern=<regexp>
EOTXT
}


if (! GetOptions(
    "f|config=s" => \$opt_f,
    "t|timeout=i" => \$opt_t,
    "V|version" => \$opt_V,
    "h|help" => \$opt_h,
    "d|debug" => \$opt_d,
    "v|verbose" => \$opt_v,
    "tag=s" => \$opt_tag,
    "logfile=s" => \$opt_logfile,
    "rotation=s" => \$opt_rotation,
    "criticalpattern=s" => \$opt_criticalpattern,
    "criticalexception=s" => \$opt_criticalexception,
    "warningpattern=s" => \$opt_warningpattern,
    "warningexception=s" => \$opt_warningexception,
    "okpattern=s" => \$opt_okpattern,
    "noprotocol" => \$opt_noprotocol,
    "syslogserver" => \$opt_syslogserver,
    "syslogclient=s" => \$opt_syslogclient,
    "sticky:s" => \$opt_sticky,
    "noperfdata" => \$opt_noperfdata,
    "searches=s" => \$opt_selectedsearches,
    "selectedsearches=s" => \$opt_selectedsearches,
    "template=s" => \$opt_template
 )) {
  print_help();
  exit UNKNOWN;
}

if ($opt_V) {
  print_version();
  exit UNKNOWN;
}
if ($opt_h) {
  print_help();
  exit UNKNOWN;
}

if (! $opt_f && ! $opt_logfile) {
  print_usage();
  exit UNKNOWN;
}
if (! $opt_selectedsearches) {
  $opt_selectedsearches = "";
}
if (my $cl = Nagios::CheckLogfiles->new({
      cfgfile => $opt_f ? $opt_f : undef,
      searches => [ map { foreach my $key (keys %{$_}) { 
      	  delete $_->{$key} unless $_->{$key}}; $_; } ({
          tag => $opt_tag ? $opt_tag : undef,
          logfile => $opt_logfile ? $opt_logfile : undef,
          rotation => $opt_rotation ? $opt_rotation : undef,
          criticalpatterns => $opt_criticalpattern ? $opt_criticalpattern : undef,
          criticalexceptions => $opt_criticalexception ? $opt_criticalexception : undef,
          warningpatterns => $opt_warningpattern ? $opt_warningpattern : undef,
          warningexceptions => $opt_warningexception ? $opt_warningexception : undef,
          okpatterns => $opt_okpattern ? $opt_okpattern : undef,
          options => join(',', grep { $_ }
              $opt_noprotocol ? "noprotocol" : undef,
              $opt_noperfdata ? "noperfdata" : undef,
              $opt_syslogserver ? "syslogserver" : undef,
              $opt_syslogclient ? "syslogclient=".$opt_syslogclient : undef,
              defined $opt_sticky ? "sticky".($opt_sticky ? "=" : "") : undef )
      })],
      selectedsearches => [split(/,/, $opt_selectedsearches)],
      dynamictag => $opt_tag ? $opt_tag : undef
  })) {
  $cl->{verbose} = $opt_v ? 1 : 0;
  $cl->{timeout} = $opt_t ? $opt_t : 60;
  $cl->run();
  printf "%s%s\n", $cl->{exitmessage}, $cl->{perfdata} ? "|".$cl->{perfdata} : "";
  exit $cl->{exitcode};
} else {
  printf "%s\n", $Nagios::CheckLogfiles::ExitMsg;
  exit $Nagios::CheckLogfiles::ExitCode;
}

